# encoding: UTF-8

#
#  mscsv.rb  --  Read Microsofts shitty CSV
#

require "supplement"

# When Microsoft's Excel exports a file as a CSV, there is hardly
# a convention that will not be violated.  Fields that contain
# newlines will be exported in quotes and the newlines will not be
# escaped.  They will be written just as one <code>"\x0a"</code>.
# Yet, the records themselves still are separated by MS newlines,
# that is <code>"\x0d\x0a"</code>.
#
class MsCsv

  VERSION = "1.0"

  class <<self

    # :call-seq:
    #   MsCsv.open( filename, sep = nil, quote = nil) { |csv| ... }  -> obj
    #
    # Open a file.  The file may only be read as I refuse to write
    # such weird formats.
    #
    # Your guess is right what the +sep+ and +quote+ parameters mean.
    #
    def open name, sep = nil, quote = nil
      File.open name do |f|
        i = new f, sep, quote
        yield i
      end
    end

  end

  SEP   = ";"
  QUOTE = '"'

  attr_accessor :encoding

  # :call-seq:
  #   MsCsv.new( file, sep = nil, quote = nil)   -> obj
  #
  # The +sep+ and +quote+ parameters are <code>";"</code> and
  # <code>"\""</code> by default.  More than one character may be
  # specified.
  #
  def initialize file, sep = nil, quote = nil
    @file = file
    @sep, @quote = sep||SEP, quote||QUOTE
  end

  # :call-seq:
  #   csv.each_record  { |ary| ... }   -> nil
  #
  # Iterate through the CSV file.  The fields will be returned as
  # strings or +nil+.  A line consists of at least one field, so
  # an empty line will yield a one-element array containing +nil+.
  #
  def each_record
    unless defined? Encoding then
      require "iconv"
      @iconv = Iconv.new "utf-8", @encoding||"ms-ansi"
    end
    while l = read_line_utf8 do
      record, field = [], ""
      until l =~ /^$/ do
        c = l.eat 1
        if    @sep.include? c   then
          record.push field
          field = ""
        elsif @quote.include? c then
          q = c
          while l.notempty? or (l = read_line_utf8) do
            c = l.eat 1
            if c == q then
              d = l.head 1
              if q == d then
                field << q
                l.eat 1
              else
                break
              end
            else
              field << c
            end
          end
        else
          field << c
        end
      end
      record.push field
      yield record
    end
  end

  # :call-seq:
  #   csv.each_notempty  { |f,g,...| ... }   -> nil
  #
  # Same as +each_record+ except that records containing only
  # +nil+ fields will be skipped.
  #
  def each_notempty
    each_record { |r|
      yield r unless r.compact.empty?
    }
  end

  # :stopdoc:

  class Str
    def initialize f ; @str = f ; end
    def val ; @str ; end
  end
  class Sstr
    def initialize f
      @str = case f
        when "#NV" then nil
        else            f.strip.notempty?
      end
    end
    def val ; @str ; end
  end
  class Num
    def initialize f
      f = f.strip
      if f =~ /\A\d{1,3}(?:\.\d{3})*/ then
        f.delete! "."
      elsif f =~ /\A\d{1,3}(?:,\d{3})*/ then
        f.delete! ","
      end
      @i = Integer f
    end
    def val ; @i ; end
  end
  class Cur
    RE_B = /\A(-?\d{0,3}(?:[,.]?\d{3})*(?:[.,](?:\d\d|-))?)\s*[^0-9.,-]+\z/
    RE_F = /\A[^0-9.,-]+\s*(-?\d{0,3}(?:[,.]?\d{3})*(?:[.,](?:\d\d|-))?)\z/
    RE_M = /\A(\d{1,3}(?:[,.]\d{3})*[.,]\d\d)(-?)\z/
    def initialize f
      f = f.strip
      @c = case f
        when RE_B, RE_F then Currency.parse $1
        when RE_M       then Currency.parse "#$2#$1"
        when ""         then Currency.parse nil
        else                 Currency.parse f
      end
    rescue
      raise "Not a valid currency value: #{f}."
    end
    def val ; @c ; end
  end
  class Dat
    def initialize f
      fmt = case f.strip
        when /\A\d\d?\.\d\d?\.\d\d\d\d\z/ then "%d.%m.%Y"
        when /\A\d\d?\.\d\d?\.\d\d\z/     then "%d.%m.%y"
        when /\A\d{4}-\d\d-\d\d\z/        then "%Y-%m-%d"
        when ""                           then return
        else
          raise "Not a valid date: #{f}"
      end
      @d = Date.strptime f, fmt rescue nil
    end
    def val ; @d ; end
  end
  class Tim
    def initialize f
      @t = Time.parse f
    end
    def val ; @t ; end
  end
  class Bool
    def initialize f
      f = f.strip
      f.downcase!
      @b = case f
        when "yes", "true",  "y", "t", "on",  /enable/  then
          true
        when "no",  "false", "n", "f", "off", /disable/ then
          false
        when "ja",   "oui", "evet"  then
          true
        when "nein", "non", "hayır" then
          false
        when /\A[01]\z/ then
          not f.to_i.zero?
        when "" then
          false
        when /\An/i, /^\Ak[^a-z]*a/ then  # "not", German "nicht", "k.A."
          false
        else
          true
      end
    end
    def val ; @b ; end
  end

  FORMATS = {
    "s" => Str,
    "S" => Sstr,
    "n" => Num,
    "c" => Cur,
    "d" => Dat,
    "t" => Tim,
    "b" => Bool,
  }

  # :startdoc:

  # :call-seq:
  #   csv.each_as( recdef) { |rec,err| ... }   -> nil
  #
  # Return record with fields converted to several types.  In case
  # any of the fields in a record may not be converted legally,
  # the +err+ variable will contain the exception and the fields will be
  # returned as strings.  At present, following types may be
  # demanded:
  #
  #   s => string
  #   S => stripped string, nil if empty
  #   n => integer
  #   c => currency   # class +Currency+ must respond to +parse+ (*)
  #   d => date       # class +Date+ must respond to +strptime+ (*)
  #   t => time       # class +Time+ must respond to +parse+ (*)
  #   b => boolean
  #
  #   (*) choose the appropriate +require+ yourself.
  #
  # Example:
  #
  #   MsCsv.open "somefile.csv" do |f|
  #     f.each_as "ndsscb" do |r,d|
  #       puts r.length.inspect + "  " + r.inspect unless d
  #     end
  #   end
  #
  def each_as recdef
    case recdef
      when String then recdef = recdef.scan /\S/
    end
    d = recdef.map { |f| FORMATS[ f] }
    each_notempty do |r|
      begin
        i = 0
        r = (d.zip r).map do |(fmt,fld)|
          i += 1
          if fld then
            f = fmt.new fld
            f.val
          end
        end
      rescue
        r = r.map { |x| x.notempty? }
        err = "#{i}: #$!"
      end
      yield r, err
    end
  end

  private

  def read_line_utf8
    l = @file.readline
    if @iconv then
      l = @iconv.iconv l
      l.gsub! /\xc2\xa0/, " "
    else
      l.force_encoding @encoding||Encoding::Windows_1252
      l.encode! Encoding::UTF_8
      l.gsub! "\u00a0", " "
    end
    l
  rescue EOFError
  end

end

